@@ -1,3 +1,5 @@
+{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE LambdaCase #-}
@@ -14,28 +16,35 @@ where
 
 import Control.Monad (MonadPlus (mplus), foldM, foldM_)
 import Control.Monad.State (MonadTrans (lift))
-import qualified Data.ByteString as BS
 import Data.List (intersperse, sortOn)
 import qualified Data.Map as M
 import qualified Data.Map.Ordered as OM
-import Data.Maybe (isJust)
+import Data.Maybe (isJust, fromMaybe)
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
 import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text as T
+import qualified Data.Text.Encoding as TE
 import qualified Data.Vector as V
-import System.FilePath (replaceFileName, takeBaseName)
+import GHC.Generics (Generic)
+import System.FilePath (replaceFileName, takeBaseName, takeDirectory, (</>))
 import Text.Parsec
 import Typst.Bind (destructuringBind)
 import Typst.Methods (getMethod)
-import Typst.Module.Standard (loadFileText, standardModule)
+import Typst.Module.Standard (loadFileText, standardModule, symModule,
+                              sysModule)
+import Typst.Module.Math (mathModule)
+import Typst.MathClass (mathClassOf, MathClass(Relation))
 import Typst.Parse (parseTypst)
 import Typst.Regex (match)
 import Typst.Show (applyShowRules)
 import Typst.Syntax
 import Typst.Types
 import Typst.Util (makeFunction, nthArg)
+import qualified Toml as Toml
+import qualified Toml.FromValue as Toml
+import qualified Toml.FromValue.Generic as Toml
 
 -- import Debug.Trace
 
@@ -43,17 +52,20 @@ import Typst.Util (makeFunction, nthArg)
 -- replacing it with content.
 evaluateTypst ::
   Monad m =>
-  -- | Function to read a file
-  (FilePath -> m BS.ByteString) ->
+  -- | Dictionary of functions for IO operations
+  Operations m ->
   -- | Path of parsed content
   FilePath ->
   -- | Markup produced by 'parseTypst'
   [Markup] ->
-  m (Either ParseError (Seq Content))
-evaluateTypst loadBytes =
+  m (Either ParseError Content)
+evaluateTypst operations =
   runParserT
-    (mconcat <$> many pContent <* eof)
-    initialEvalState {evalLoadBytes = loadBytes}
+    (do contents <- mconcat <$> many pContent <* eof
+        -- "All documents are automatically wrapped in a document element."
+        pure $ Elt "document" Nothing [("body", VContent contents)])
+    initialEvalState { evalOperations = operations,
+                       evalPackageRoot = "." }
 
 initialEvalState :: EvalState m
 initialEvalState =
@@ -229,10 +241,9 @@ pElt = do
                 [("level", VInteger (fromIntegral level))]
           }
     Equation display ms -> inBlock BlockScope $ do
-      VModule _ mathmod <- lookupIdentifier "math"
-      importModule mathmod
-      VModule _ symmod <- lookupIdentifier "sym"
-      importModule symmod
+      importModule mathModule
+      importModule symModule
+      importModule sysModule
       oldMath <- evalMath <$> getState
       updateState $ \st -> st {evalMath = True}
       content <- pInnerContents ms
@@ -259,7 +270,18 @@ pElt = do
             named = OM.empty
           }
     MAttach mbBottomExp mbTopExp baseExp -> do
-      base <- pInnerContents [baseExp]
+      base' <- pInnerContents [baseExp]
+      pos <- getPosition
+      let base =
+            case base' of
+              [Elt "text" mbpos [("body", VContent [Txt t])]]
+                     | T.all ((== Relation) . mathClassOf ) t
+                -> [Elt "math.limits" mbpos
+                        [("body", VContent base')]]
+              [Txt t] | T.all ((== Relation) . mathClassOf ) t
+                -> [Elt "math.limits" (Just pos)
+                        [("body", VContent base')]]
+              _ -> base'
       mbBottom <-
         maybe
           (pure Nothing)
@@ -439,13 +461,38 @@ evalExpr expr =
             )
             (VNone, False)
             exprs
-    Array e -> VArray . V.fromList <$> mapM evalExpr e
+    Array es -> VArray <$> foldM
+           ( \xs x ->
+               case x of
+                 Spr y -> do
+                   val <- evalExpr y
+                   case val of
+                     VArray ys -> pure (xs <> ys)
+                     _ -> fail $ "Could not spread " <> show (valType val) <>
+                                 " into array"
+                 Reg e -> do
+                   val <- evalExpr e
+                   pure (V.snoc xs val ) )
+           []
+           es
     Dict items ->
       VDict
         <$> foldM
-          ( \m (k, e) -> do
-              val <- evalExpr e
-              pure $ m OM.|> (k, val)
+          ( \m v -> do
+              case v of
+                Reg (k, e) -> do
+                  k' <- case k of
+                           Ident i -> pure i
+                           _ -> do VString s <- evalExpr k
+                                   pure $ Identifier s
+                  val <- evalExpr e
+                  pure $ m OM.|> (k', val)
+                Spr y -> do
+                  val <- evalExpr y
+                  case val of
+                    VDict m' -> pure (m' OM.|<> m)
+                    _ -> fail $ "Could not spread " <> show (valType val) <>
+                                " into dictionary"
           )
           OM.empty
           items
@@ -590,12 +637,18 @@ evalExpr expr =
             VString t' -> pure $ VBoolean $ t' `T.isInfixOf` t
             VRegex re -> pure $ VBoolean $ match re t
             _ -> fail $ "Can't apply 'in' to " <> show v1 <> " and string"
+        VType ty ->
+          case v1 of
+            VString t' -> pure $ VBoolean $ t' `T.isInfixOf` (prettyType ty)
+            VRegex re -> pure $ VBoolean $ match re (prettyType ty)
+            _ -> fail $ "Can't apply 'in' to " <> show v1
         VArray vec -> pure $ VBoolean $ V.elem v1 vec
         VDict m ->
           case v1 of
             VString t -> pure $ VBoolean $ isJust $ OM.lookup (Identifier t) m
             _ -> pure $ VBoolean False
-        _ -> fail $ "Can't apply 'in' to " <> show v2
+        _ -> fail $ "Can't apply 'in' to " <> show v2 <> show (e1,e2)
+
     Negated e -> do
       v <- evalExpr e
       case maybeNegate v of
@@ -780,27 +833,30 @@ evalExpr expr =
       argval <- evalExpr e
       (modid, modmap) <-
         case argval of
-          VString t -> loadModule t
+          VString t -> snd <$> loadModule t
           VModule i m -> pure (i, m)
           VFunction (Just i) m _ -> pure (i, m)
           VFunction Nothing m _ -> pure ("anonymous", m)
           _ -> fail "Import requires a path or module or function"
       case imports of
         AllIdentifiers -> importModule modmap
-        SomeIdentifiers idents -> do
-          let addFromModule m ident =
+        SomeIdentifiers pairs -> do
+          let addFromModule m (ident, mbAs) =
                 case M.lookup ident modmap of
                   Nothing -> fail $ show ident <> " not defined in module"
-                  Just v -> pure $ M.insert ident v m
-          foldM addFromModule mempty idents >>= importModule
-        NoIdentifiers -> addIdentifier modid (VModule modid modmap)
+                  Just v -> pure $ M.insert (fromMaybe ident mbAs) v m
+          foldM addFromModule mempty pairs >>= importModule
+        NoIdentifiers mbAs -> do
+          let ident = fromMaybe modid mbAs
+          addIdentifier ident (VModule ident modmap)
       pure VNone
     Include e -> do
       argval <- evalExpr e
       case argval of
-        VString t -> loadModule t >>= importModule . snd
+        VString t -> do
+          (cs, _) <- loadModule t
+          pure $ VContent cs
         _ -> fail "Include requires a path"
-      pure VNone
 
 toFunction ::
   Monad m =>
@@ -862,31 +918,119 @@ toFunction mbname params e = do
           pure res
   pure fn
 
-loadModule :: Monad m => Text -> MP m (Identifier, M.Map Identifier Val)
+-- | Type of TOML configuration file used in 'findPackageEntryPoint' saved as `typst.toml`
+data Config = Config {
+  package :: PackageConfig
+
+} deriving (Show, Generic)
+
+data PackageConfig = PackageConfig {
+  entrypoint :: String
+} deriving (Show, Generic)
+
+-- | Derived generically from record field names
+instance Toml.FromValue Config where
+  fromValue = Toml.parseTableFromValue Toml.genericParseTable
+
+-- | Derived generically from record field names
+instance Toml.FromValue PackageConfig where
+  fromValue = Toml.parseTableFromValue Toml.genericParseTable
+
+findPackageEntryPoint :: Monad m => Text -> MP m FilePath
+findPackageEntryPoint modname = do
+  let (namespace, rest) = break (=='/') (drop 1 $ T.unpack modname)
+  let (name, rest') = break (==':') $ drop 1 rest
+  let version = drop 1 rest'
+  operations <- evalOperations <$> getState
+  let getEnv var = do
+        mbv <- lift $ lookupEnvVar operations var
+        case mbv of
+          Just v -> pure v
+          Nothing -> fail (var <> " not defined")
+#ifdef __MACOS__
+  homeDir <- getEnv "HOME"
+  let localDir = homeDir </> "Library" </> "Application Support" </> "typst"
+  let cacheDir = homeDir </> "Library" </> "Caches" </> "typst"
+#elif __WINDOWS__
+  appDataDir <- getEnv "APPDATA"
+  let localDir = appDataDir </> "typst"
+  localAppDataDir <- getEnv "LOCALAPPDATA"
+  let cacheDir = localAppDataDir </> "typst"
+#else
+  homeDir <- getEnv "HOME"
+  dataDir <- lift (lookupEnvVar operations "XDG_DATA_HOME") >>=
+               maybe (pure (homeDir </> ".local" </> "share")) pure
+  cacheDir' <- lift (lookupEnvVar operations "XDG_CACHE_HOME") >>=
+               maybe (pure (homeDir </> ".cache")) pure
+  let localDir = dataDir </> "typst"
+  let cacheDir = cacheDir' </> "typst"
+#endif
+  let subpath = "packages" </> namespace </> name </> version
+  inLocal <- lift $ checkExistence operations (localDir </> subpath </> "typst.toml")
+  tomlPath <-
+     if inLocal
+        then pure (localDir </> subpath </> "typst.toml")
+        else do
+          inCache <- lift $ checkExistence operations (cacheDir </> subpath </> "typst.toml")
+          if inCache
+             then pure (cacheDir </> subpath </> "typst.toml")
+             else fail $ "Could not find package in local packages or cache. Looked in\n" ++
+                    (localDir </> subpath) ++ "\n" ++ (cacheDir </> subpath) ++
+                    "\nCompile with typst compile to bring the package into your local cache."
+             -- TODO? fetch from CDN if not present in cache?
+  tomlString <- T.unpack . TE.decodeUtf8 <$> lift (loadBytes operations tomlPath)
+  case Toml.decode tomlString of
+    Toml.Failure e -> fail (unlines ("Failure loading typst.toml" : e))
+    Toml.Success _warnings cfg -> -- ignores warnings like unused keys in TOML
+      pure (replaceFileName tomlPath (entrypoint (package cfg)))
+
+loadModule :: Monad m => Text
+           -> MP m (Seq Content, (Identifier, M.Map Identifier Val))
 loadModule modname = do
   pos <- getPosition
-  let fp = replaceFileName (sourceName pos) (T.unpack modname)
-  let modid = Identifier (T.pack $ takeBaseName fp)
+  (fp, modid, mbPackageRoot) <-
+        if T.take 1 modname == "@"
+           then do
+            fp' <- findPackageEntryPoint modname
+            pure (fp',
+                   Identifier
+                    (T.pack $ takeWhile (/= ':') . takeBaseName $
+                       T.unpack modname),
+                   Just (takeDirectory fp'))
+           else if T.take 1 modname == "/" -- refers to path relative to package root
+                then do
+                  packageRoot <- evalPackageRoot <$> getState
+                  pure (packageRoot </> drop 1 (T.unpack modname),
+                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
+                        Nothing)
+                else pure (replaceFileName (sourceName pos) (T.unpack modname),
+                        Identifier (T.pack $ takeBaseName $ T.unpack modname),
+                        Nothing)
   txt <- loadFileText fp
   case parseTypst fp txt of
     Left err -> fail $ show err
     Right ms -> do
-      loadBytes <- evalLoadBytes <$> getState
+      operations <- evalOperations <$> getState
       res <-
         lift $
           runParserT
-            ( inBlock BlockScope $ -- add new identifiers list
-                many pContent *> eof *> getState
+            ( inBlock BlockScope $ do -- add new identifiers list
+                cs <- mconcat <$> many pContent
+                eof
+                s <- getState
+                pure (cs, s)
             )
-            initialEvalState {evalLoadBytes = loadBytes}
+            initialEvalState{evalOperations = operations,
+                             evalPackageRoot = fromMaybe (evalPackageRoot initialEvalState)
+                                                   mbPackageRoot }
             fp
             ms
       case res of
         Left err' -> fail $ show err'
-        Right st ->
+        Right (contents, st) ->
           case evalIdentifiers st of
             [] -> fail "Empty evalIdentifiers in module!"
-            ((_, m) : _) -> pure (modid, m)
+            ((_, m) : _) -> pure (contents, (modid, m))
 
 importModule :: Monad m => M.Map Identifier Val -> MP m ()
 importModule m = updateState $ \st ->
@@ -992,57 +1136,60 @@ inBlock scope pa = do
   pure result
 
 updateExpression :: Monad m => Expr -> Val -> MP m ()
-updateExpression e val =
-  case e of
-    Ident i -> updateIdentifier i val
-    FuncCall
-      (FieldAccess (Ident (Identifier "at")) e')
-      [NormalArg (Literal (Int idx))] ->
-        do
-          ival <- evalExpr e'
-          case ival of
-            VArray v ->
-              updateExpression e' $ VArray $ v V.// [(fromIntegral idx, val)]
-            _ -> fail $ "Cannot update expression " <> show e
-    FuncCall (FieldAccess (Ident (Identifier "first")) e') [] ->
-      updateExpression
-        ( FuncCall
-            (FieldAccess (Ident (Identifier "at")) e')
-            [NormalArg (Literal (Int 0))]
-        )
-        val
-    FuncCall (FieldAccess (Ident (Identifier "last")) e') [] ->
-      updateExpression
-        ( FuncCall
-            (FieldAccess (Ident (Identifier "at")) e')
-            [NormalArg (Literal (Int (-1)))]
-        )
-        val
-    FuncCall
-      (FieldAccess (Ident (Identifier "at")) e')
-      [NormalArg (Literal (String fld))] ->
-        do
-          ival <- evalExpr e'
-          case ival of
-            VDict d ->
-              updateExpression e' $
-                VDict $
-                  OM.alter
-                    ( \case
-                        Just _ -> Just val
-                        Nothing -> Just val
-                    )
-                    (Identifier fld)
-                    d
-            _ -> fail $ "Cannot update expression " <> show e
-    FieldAccess (Ident (Identifier fld)) e' ->
-      updateExpression
-        ( FuncCall
-            (FieldAccess (Ident (Identifier "at")) e')
-            [NormalArg (Literal (String fld))]
-        )
-        val
-    _ -> fail $ "Cannot update expression " <> show e
+updateExpression
+  (FuncCall (FieldAccess (Ident (Identifier "first")) e') []) val
+  = updateExpression ( FuncCall
+                       (FieldAccess (Ident (Identifier "at")) e')
+                       [NormalArg (Literal (Int 0))]
+                     ) val
+updateExpression
+  (FuncCall (FieldAccess (Ident (Identifier "last")) e') []) val
+  = updateExpression ( FuncCall
+                       (FieldAccess (Ident (Identifier "at")) e')
+                       [NormalArg (Literal (Int (-1)))]
+                     ) val
+updateExpression
+  (FieldAccess (Ident (Identifier fld)) e') val
+  = updateExpression' True e' (Literal (String fld)) val -- see #26
+updateExpression (Ident i) val = updateIdentifier i val
+updateExpression
+  (FuncCall
+      (FieldAccess (Ident (Identifier "at")) e') [NormalArg arg]) val
+  = updateExpression' False e' arg val
+updateExpression e _ = fail $ "Cannot update expression " <> show e
+
+updateExpression' :: Monad m => Bool -> Expr -> Expr -> Val -> MP m ()
+updateExpression' allowNewIndices e arg val = do
+    container <- evalExpr e
+    idx <- evalExpr arg
+    case container of
+      VArray v ->
+        case idx of
+          VInteger i ->
+            let i' = fromIntegral i
+            in case v V.!? i' of
+                 Nothing | not allowNewIndices
+                     -> fail $ "Vector does not contain index " <> show i'
+                 _ -> updateExpression e $ VArray $ v V.// [(i', val)]
+          _ -> fail $ "Cannot index array with " <> show idx
+      VDict d ->
+        case idx of
+          VString fld ->
+            case OM.lookup (Identifier fld) d of
+              Nothing | not allowNewIndices
+                     -> fail $ "Dictionary does not contain key " <> show fld
+              _ -> updateExpression e $
+                           VDict $
+                             OM.alter
+                               ( \case
+                                   Just _ -> Just val
+                                   Nothing -> Just val
+                               )
+                               (Identifier fld)
+                               d
+          _ -> fail $ "Cannot index dictionary with " <> show idx
+      _ -> fail $ "Cannot update expression " <> show e
+
 
 toSelector :: Monad m => Val -> MP m Selector
 toSelector (VSelector s) = pure s
