@@ -7,24 +7,25 @@
 
 module Typst.Methods
   ( getMethod,
-    applyPureFunction,
     formatNumber,
+    applyPureFunction
   )
 where
 
-import Control.Monad (MonadPlus (mplus), foldM)
+import Control.Monad (MonadPlus (mplus), foldM, void)
 import Control.Monad.Reader (MonadReader (ask), MonadTrans (lift))
 import qualified Data.Array as Array
 import qualified Data.Foldable as F
 import Data.List (intersperse, sort, sortOn)
 import qualified Data.Map as M
 import qualified Data.Map.Ordered as OM
-import Data.Maybe (fromMaybe)
+import Data.Maybe (fromMaybe, listToMaybe)
 import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Vector as V
-import Text.Parsec (getState, runParserT, updateState)
-import Typst.Module.Standard (standardModule)
+import Text.Parsec
+import Text.Parsec.String (Parser)
+import Typst.Module.Standard (applyPureFunction)
 import Typst.Regex
   ( RE (..),
     RegexMatch (..),
@@ -37,6 +38,7 @@ import Typst.Regex
   )
 import Typst.Types
 import Typst.Util (allArgs, makeFunction, namedArg, nthArg)
+import Data.Time (toGregorian, dayOfWeek, formatTime, defaultTimeLocale, UTCTime(..))
 
 -- import Debug.Trace
 
@@ -65,16 +67,16 @@ getMethod updateVal val fld = do
         "at" ->
           pure $ makeFunction $ do
             key <- nthArg 1
-            defval <- namedArg "default" `mplus` pure VNone
+            defval <- namedArg "default" VNone
             case OM.lookup (Identifier key) m of
               Nothing -> pure defval
               Just v -> pure v
-        "insert" ->
+        "insert" -> do
           pure $ makeFunction $ do
             key <- nthArg 1
             v <- nthArg 2
             lift $ updateVal $ VDict $ m OM.|> (Identifier key, v)
-            pure v
+            pure VNone
         "keys" ->
           pure $
             makeFunction $
@@ -144,6 +146,8 @@ getMethod updateVal val fld = do
       case fld of
         "len" ->
           pure $ makeFunction $ pure $ VInteger (fromIntegral $ T.length t)
+        "rev" ->
+          pure $ makeFunction $ pure $ VString (T.reverse t)
         "first" ->
           if T.null t
             then fail "string is empty"
@@ -159,10 +163,9 @@ getMethod updateVal val fld = do
         "slice" ->
           pure $ makeFunction $ do
             start <- toPos <$> nthArg 1
-            end <-
-              (toPos <$> nthArg 2)
-                `mplus` ((+ start) <$> namedArg "count")
-                `mplus` pure (T.length t)
+            mbcount <- namedArg "count" Nothing
+            end <- (toPos <$> nthArg 2) `mplus`
+                      pure (maybe (T.length t) (+ start) mbcount)
             if end < start
               then pure $ VString ""
               else pure $ VString $ T.take (end - start) $ T.drop start t
@@ -226,13 +229,15 @@ getMethod updateVal val fld = do
         "replace" -> pure $ makeFunction $ do
           patt :: RE <- nthArg 1
           (replacement :: Val) <- nthArg 2
-          mbCount :: Maybe Int <- namedArg "count" `mplus` pure Nothing
+          mbCount :: Maybe Int <- namedArg "count" Nothing
           case mbCount of
             Just 0 -> pure $ VString t
             _ ->
               case replacement of
                 VString r ->
                   pure $ VString $ replaceRegex patt mbCount (const r) t
+                VSymbol (Symbol r _ _) ->
+                  pure $ VString $ replaceRegex patt mbCount (const r) t
                 VFunction _ _ f ->
                   pure $
                     VString $
@@ -257,8 +262,8 @@ getMethod updateVal val fld = do
                 _ -> fail "replacement must be string or function"
         "trim" -> pure $ makeFunction $ do
           (RE patt _) <- nthArg 1 `mplus` makeRE "[[:space:]]*"
-          (repeated :: Bool) <- namedArg "repeat" `mplus` pure True
-          (mbAt :: Maybe Val) <- namedArg "at" `mplus` pure Nothing
+          (repeated :: Bool) <- namedArg "repeat" True
+          (mbAt :: Maybe Val) <- namedArg "at" Nothing
           let patt' =
                 if repeated
                   then "(" <> patt <> ")*"
@@ -336,7 +341,7 @@ getMethod updateVal val fld = do
                   <> T.unpack (repr (VContent cs))
         "at" -> pure $ makeFunction $ do
           (field :: Text) <- ask >>= getPositionalArg 1 >>= fromVal
-          defval <- namedArg "default" `mplus` pure VNone
+          defval <- namedArg "default" VNone
           case F.toList cs of
             [Elt _ _ fields] ->
               case M.lookup (Identifier field) fields of
@@ -351,6 +356,11 @@ getMethod updateVal val fld = do
             [Elt _ _ fields]
               | Just x <- M.lookup "text" fields -> pure x
             _ -> fail "Content is not a single text element"
+        "fields" -> pure $ makeFunction $ do
+          VDict <$>
+            case F.toList cs of
+              (Elt _ _ fields:_) -> pure $ OM.fromList $ M.toList fields
+              _ -> pure OM.empty
         _ ->
           let childrenOrFallback =
                 if fld == "children"
@@ -370,6 +380,12 @@ getMethod updateVal val fld = do
         "term" -> pure $ VContent t
         "description" -> pure $ VContent d
         _ -> noMethod "TermItem" fld
+    VVersion xs ->
+      case fld of
+        "at" -> pure $ makeFunction $ do
+                  i <- nthArg 1
+                  pure $ VInteger $ fromMaybe 0 $ listToMaybe $ drop i xs
+        _ -> noMethod "Version" fld
     VArray v -> do
       let toPos n =
             if n < 0
@@ -392,7 +408,7 @@ getMethod updateVal val fld = do
                 else pure $ V.last v
         "at" -> pure $ makeFunction $ do
           pos <- toPos <$> nthArg 1
-          defval <- namedArg "default" `mplus` pure VNone
+          defval <- namedArg "default" VNone
           pure $ fromMaybe defval $ v V.!? pos
         "push" -> pure $ makeFunction $ do
           x <- nthArg 1
@@ -408,10 +424,9 @@ getMethod updateVal val fld = do
                   pure $ V.last v
         "slice" -> pure $ makeFunction $ do
           start <- toPos <$> nthArg 1
-          end <-
-            (toPos <$> nthArg 2)
-              `mplus` ((+ start) <$> namedArg "count")
-              `mplus` pure (V.length v)
+          mbcount <- namedArg "count" Nothing
+          end <- (toPos <$> nthArg 2) `mplus`
+                    pure (maybe (V.length v) (+ start) mbcount)
           if V.length v < end
             then fail "array contains insufficient elements for slice"
             else
@@ -421,9 +436,14 @@ getMethod updateVal val fld = do
         "split" -> pure $ makeFunction $ do
           spliton <- nthArg 1
           let go v' = case V.break (== spliton) v' of
-                (a, b) | V.null b -> if V.null a then [] else [VArray a]
+                (a, b) | V.null b -> [VArray a | not (V.null a)]
                 (a, b) -> VArray a : go (V.drop 1 b)
           pure $ VArray $ V.fromList $ go v
+        "intersperse" -> pure $ makeFunction $ do
+          sep <- nthArg 1
+          pure $ VArray . V.fromList . intersperse sep . V.toList $ v
+        "dedup" -> pure $ makeFunction $ do
+          pure $ VArray $ deduplicateVector v
         "insert" -> pure $ makeFunction $ do
           pos <- toPos <$> nthArg 1
           newval <- nthArg 2
@@ -522,14 +542,14 @@ getMethod updateVal val fld = do
         "rev" -> pure $ makeFunction $ pure $ VArray $ V.reverse v
         "join" -> pure $ makeFunction $ do
           separator <- nthArg 1
-          lastsep <- namedArg "last" `mplus` pure separator
+          lastsep <- namedArg "last" separator
           let xs' = F.toList v
           let xs = case xs' of
                 [] -> []
                 _ -> intersperse separator (init xs') ++ [lastsep, last xs']
           foldM joinVals VNone xs
         "sorted" -> pure $ makeFunction $ do
-          (mbKeyFn :: Maybe Function) <- namedArg "key" `mplus` pure Nothing
+          (mbKeyFn :: Maybe Function) <- namedArg "key" Nothing
           case mbKeyFn of
             Nothing -> pure $ VArray $ V.fromList $ sort $ V.toList v
             Just (Function kf) -> do
@@ -537,10 +557,17 @@ getMethod updateVal val fld = do
               VArray . V.fromList . map fst . sortOn snd
                 <$> (mapM (\x -> (x,) <$> kf' x) (V.toList v))
         "zip" -> pure $ makeFunction $ do
-          (v' :: V.Vector Val) <- ask >>= getPositionalArg 1
-          pure $ VArray $ V.map pairToArray $ V.zip v v'
+          (xs :: [Val]) <- positional <$> ask
+          let len = V.length v
+          pure $ VArray $ V.filter (/= VNone) $
+            V.map (\i -> maybe VNone (VArray . V.fromList)
+                           (mapM (\x ->
+                              case x of
+                                VArray v' -> v' V.!? i
+                                _ -> Nothing) (val : xs)))
+              (V.enumFromTo 0 (len - 1))
         "sum" -> pure $ makeFunction $ do
-          mbv <- namedArg "default" `mplus` pure Nothing
+          mbv <- namedArg "default" Nothing
           case V.uncons v of
             Nothing ->
               maybe
@@ -558,7 +585,7 @@ getMethod updateVal val fld = do
                     (Just h)
                     rest
         "product" -> pure $ makeFunction $ do
-          mbv <- namedArg "default" `mplus` pure Nothing
+          mbv <- namedArg "default" Nothing
           case V.uncons v of
             Nothing ->
               maybe
@@ -613,23 +640,55 @@ getMethod updateVal val fld = do
         "pos" -> pure $ makeFunction $ pure $ VArray $ V.fromList (positional args)
         "named" -> pure $ makeFunction $ pure $ VDict $ named args
         _ -> noMethod "Arguments" fld
+    VDateTime mbdate mbtime -> do
+      let toSeconds = (floor :: Double -> Integer) . realToFrac
+      case fld of
+        "year" -> pure $ makeFunction $
+          pure $ case toGregorian <$> mbdate of
+                   Nothing -> VNone
+                   Just (y,_,_) -> VInteger (fromIntegral y)
+        "month" -> pure $ makeFunction $
+          pure $ case toGregorian <$> mbdate of
+                   Nothing -> VNone
+                   Just (_,m,_) -> VInteger (fromIntegral m)
+        "day" -> pure $ makeFunction $
+          pure $ case toGregorian <$> mbdate of
+                   Nothing -> VNone
+                   Just (_,_,d) -> VInteger (fromIntegral d)
+        "weekday" -> pure $ makeFunction $
+          pure $ case dayOfWeek <$> mbdate of
+                   Nothing -> VNone
+                   Just d-> VInteger (fromIntegral $ fromEnum d)
+        "hour" -> pure $ makeFunction $
+          pure $ case toSeconds <$> mbtime of
+            Nothing -> VNone
+            Just t -> VInteger $ t `div` 3600
+        "minute" -> pure $ makeFunction $
+          pure $ case toSeconds <$> mbtime of
+            Nothing -> VNone
+            Just t -> VInteger $ (t `mod` 3600) `div` 60
+        "second" -> pure $ makeFunction $
+          pure $ case toSeconds <$> mbtime of
+            Nothing -> VNone
+            Just t -> VInteger $ t `mod` 60
+        "display" -> pure $ makeFunction $ do
+          mbfmt <- nthArg 1 `mplus` pure Nothing
+          mbformat <- case mbfmt of
+            Nothing -> pure Nothing
+            Just fmt ->
+              case toTimeFormat <$> parseDisplayFormat fmt of
+                Left e -> fail $ "Could not parse display format: " <> show e
+                Right f -> pure $ Just f
+          pure $ VString $ T.pack $
+            case (mbdate, mbtime) of
+              (Nothing, Just t) -> formatTime defaultTimeLocale (fromMaybe "%X" mbformat) t
+              (Just d, Nothing) -> formatTime defaultTimeLocale (fromMaybe "%F" mbformat) d
+              (Nothing, Nothing) -> ""
+              (Just d, Just t) -> formatTime defaultTimeLocale (fromMaybe "%X %F" mbformat)
+                                    (UTCTime d t)
+        _ -> noMethod "DateTime" fld
     _ -> noMethod (drop 1 $ takeWhile (/= ' ') $ show val) fld
 
-pairToArray :: (Val, Val) -> Val
-pairToArray (x, y) = VArray $ V.fromList [x, y]
-
-applyPureFunction :: Function -> [Val] -> Attempt Val
-applyPureFunction (Function f) vals =
-  let args = Arguments vals OM.empty
-   in case runParserT (f args) initialEvalState "" [] of
-        Failure s -> Failure s
-        Success (Left s) -> Failure $ show s
-        Success (Right v) -> Success v
-
-initialEvalState :: MonadFail m => EvalState m
-initialEvalState =
-  emptyEvalState { evalIdentifiers = [(BlockScope, standardModule)] }
-
 formatNumber :: Text -> Int -> Text
 formatNumber t n = F.foldMap go $ T.unpack t
   where
@@ -664,3 +723,107 @@ toRomanNumeral x
   | x == 4 = "IV"
   | x >= 1 = "I" <> toRomanNumeral (x - 1)
   | otherwise = ""
+
+-- parser for DateTime display format
+
+data FormatPart =
+    Literal String
+  | Variable String [(String, String)]
+  deriving Show
+
+parseDisplayFormat :: String -> Either ParseError [FormatPart]
+parseDisplayFormat = parse (many pFormatPart <* eof) ""
+
+pFormatPart :: Parser FormatPart
+pFormatPart = pVariable <|> pLiteral
+
+pLiteral :: Parser FormatPart
+pLiteral = Literal <$> many1 (satisfy (/='['))
+
+pVariable :: Parser FormatPart
+pVariable = do
+  void $ char '['
+  name <- many1 letter
+  spaces
+  modifiers <- many pModifier
+  void $ char ']'
+  pure $ Variable name modifiers
+
+pModifier :: Parser (String, String)
+pModifier = do
+  name <- many1 letter
+  void $ char ':'
+  spaces
+  val <- many1 alphaNum
+  spaces
+  pure (name, val)
+
+-- convert formatparts into Data.Time format string
+
+toTimeFormat :: [FormatPart] -> String
+toTimeFormat = concatMap toTimeFormatPart
+
+toTimeFormatPart :: FormatPart -> String
+toTimeFormatPart (Literal s) = foldr esc "" s
+ where
+  esc '%' = ("%%" ++)
+  esc '\t' = ("%t" ++)
+  esc '\n' = ("%n" ++)
+  esc c = (c:)
+toTimeFormatPart (Variable "year" mods) =
+  withPadding mods $
+    case lookup "repr" mods of
+       Just "last_two" -> "y"
+       _ -> "Y"
+toTimeFormatPart (Variable "month" mods) =
+  withPadding mods $
+    case lookup "repr" mods of
+       Just "numerical" -> "%m"
+       Just "long" -> "b"
+       Just "short" -> "h"
+       _ -> "m"
+toTimeFormatPart (Variable "day" mods) =
+  case lookup "padding" mods of
+    Just "space" -> "%e"
+    Just "zero" -> "%d"
+    _ -> "%e"
+toTimeFormatPart (Variable "week_number" mods) =
+  withPadding mods $
+    case lookup "repr" mods of
+      Just "ISO" -> "V"
+      Just "sunday" -> "U"
+      Just "monday" -> "W"
+      _ -> "V"
+toTimeFormatPart (Variable "weekday" mods) =
+  withPadding mods $
+     case lookup "repr" mods of
+      Just "long" -> "A"
+      Just "short" -> "a"
+      Just "sunday" -> "w"
+      Just "monday" -> "u"
+      _ -> ""
+toTimeFormatPart (Variable "hour" mods) =
+  case lookup "hour" mods of
+    Just "24" | lookup "padding" mods == Just "zero" -> "%H"
+              | otherwise -> "%k"
+    Just "12" | lookup "padding" mods == Just "zero" -> "%I"
+              | otherwise -> "%l"
+    _ -> "%k"
+toTimeFormatPart (Variable "period" mods) =
+  case lookup "case" mods of
+    Just "lower" -> "%P"
+    _ -> "%p"
+toTimeFormatPart (Variable "minute" _) = "%M"
+toTimeFormatPart (Variable "second" _) = "%S"
+toTimeFormatPart _ = "?"
+
+withPadding :: [(String, String)] -> String -> String
+withPadding mods s = '%' :
+  case lookup "padding" mods of
+       Just "zero" -> '0' : s
+       Just "space" -> '_' : s
+       _ -> s
+
+deduplicateVector :: Eq a => V.Vector a -> V.Vector a
+deduplicateVector =
+  V.foldl' (\acc x -> if x `V.elem` acc then acc else acc `V.snoc` x) mempty
