@@ -6,13 +6,17 @@
 
 module Typst.Module.Standard
   ( standardModule,
+    symModule,
+    sysModule,
     loadFileText,
+    applyPureFunction
   )
 where
 
+import Data.Char (ord, chr)
 import Control.Applicative ((<|>))
 import Control.Monad (mplus, unless)
-import Control.Monad.Reader (lift)
+import Control.Monad.Reader (lift, asks)
 import qualified Data.Aeson as Aeson
 import qualified Data.ByteString.Lazy as BL
 import qualified Data.Csv as Csv
@@ -27,9 +31,10 @@ import qualified Data.Text as T
 import qualified Data.Text.Encoding as TE
 import qualified Data.Vector as V
 import qualified Data.Yaml as Yaml
-import Text.Parsec (getPosition, getState, updateState)
+import Text.Parsec (getPosition, getState, updateState, runParserT)
 import Text.Read (readMaybe)
 import qualified Text.XML as XML
+import qualified Toml
 import Typst.Emoji (typstEmojis)
 import Typst.Module.Calc (calcModule)
 import Typst.Module.Math (mathModule)
@@ -37,12 +42,18 @@ import Typst.Regex (makeRE)
 import Typst.Symbols (typstSymbols)
 import Typst.Types
 import Typst.Util
+import System.FilePath ((</>))
+import Data.List (genericTake)
+import Data.Time (UTCTime(..))
+import Data.Time.Calendar (fromGregorianValid)
+import Data.Time.Clock (secondsToDiffTime)
 
 standardModule :: M.Map Identifier Val
 standardModule =
   M.fromList $
     [ ("math", VModule "math" mathModule),
       ("sym", VModule "sym" symModule),
+      ("sys", VModule "sys" sysModule),
       ("emoji", VModule "emoji" emojiModule),
       ("calc", VModule "calc" calcModule)
     ]
@@ -55,8 +66,13 @@ standardModule =
       ++ meta
       ++ foundations
       ++ construct
+      ++ time
       ++ dataLoading
 
+sysModule :: M.Map Identifier Val
+sysModule =
+  M.fromList [ makeElement (Just "sys") "version" [] ]
+
 symModule :: M.Map Identifier Val
 symModule = M.map VSymbol $ makeSymbolMap typstSymbols
 
@@ -81,6 +97,7 @@ textual =
     makeElement Nothing "smallcaps" [("body", One TContent)],
     makeElement Nothing "underline" [("body", One TContent)],
     makeElement Nothing "overline" [("body", One TContent)],
+    makeElement Nothing "highlight" [("body", One TContent)],
     makeElement Nothing "raw" [("text", One TString)],
     makeElement Nothing "smartquote" [],
     makeElement Nothing "lower" [("text", One (TString :|: TContent))],
@@ -196,15 +213,11 @@ visualize =
 meta :: [(Identifier, Val)]
 meta =
   [ makeElement Nothing "bibliography" [("path", One (TString :|: TArray))],
-    makeElement
-      Nothing
-      "cite"
-      [ ("keys", Many TString),
-        ("supplement", One (TContent :|: TNone))
-      ],
+    makeElement Nothing "cite" [("key", One TLabel)],
     makeElement Nothing "document" [],
     makeElement Nothing "figure" [("body", One TContent)],
     makeElement Nothing "heading" [("body", One TContent)],
+    makeElement Nothing "quote" [("body", One TContent)],
     makeElement Nothing "layout" [("func", One TFunction)],
     makeElement
       Nothing
@@ -219,21 +232,33 @@ meta =
       [ ("numbering", One (TString :|: TFunction)),
         ("numbers", Many TInteger)
       ],
-    makeElement Nothing "outline" [],
+    makeElementWithScope Nothing "outline"
+      []
+      [makeElement (Just "outline") "entry"
+        [("level", One TInteger),
+         ("element", One TContent),
+         ("body", One TContent),
+         ("fill", One (TContent :|: TNone)),
+         ("page", One TContent)]],
     makeElement
       Nothing
       "query"
       [ ("target", One (TLabel :|: TFunction)),
         ("location", One TLocation)
       ],
+    makeElement Nothing "metadata" [ ("value", One TAny) ],
     makeElement Nothing "ref" [("target", One TLabel)],
     makeElement Nothing "state" [("key", One TString), ("init", One TAny)],
-    makeElement Nothing "style" [("func", One TFunction)],
     makeElementWithScope
       Nothing
       "footnote"
       [("body", One TContent)]
-      [makeElement (Just "footnote") "entry" [("note", One TContent)]]
+      [makeElement (Just "footnote") "entry" [("note", One TContent)]],
+    ("style", makeFunction $ do
+        Function f <- nthArg 1
+        case applyPureFunction (Function f) [VStyles] of
+          Success x -> pure x
+          Failure e -> fail e)
   ]
 
 colors :: [(Identifier, Val)]
@@ -285,7 +310,7 @@ foundations =
         ( do
             (cond :: Bool) <- nthArg 1
             unless cond $ do
-              (msg :: String) <- namedArg "message" <|> pure "Assertion failed"
+              (msg :: String) <- namedArg "message" "Assertion failed"
               fail msg
             pure VNone
         )
@@ -305,20 +330,13 @@ foundations =
           )
         ]
     ),
-    ("panic", makeFunction $ allArgs >>= fail . unlines . map show),
+    ("panic", makeFunction $ allArgs >>= fail . T.unpack .
+                 (("panicked with: " <>) . T.unlines . map repr)),
     ("repr", makeFunction $ nthArg 1 >>= pure . VString . repr),
     ( "type",
       makeFunction $ do
         (x :: Val) <- nthArg 1
-        pure $
-          VString $
-            case valType x of
-              TAlignment ->
-                case x of
-                  VAlignment (Just _) (Just _) -> "2d alignment"
-                  _ -> "alignment"
-              TDict -> "dictionary"
-              ty -> T.toLower . T.drop 1 . T.pack . show $ ty
+        pure $ VType $ valType x
     )
   ]
 
@@ -344,8 +362,8 @@ construct =
     ( "range",
       makeFunction $ do
         first <- nthArg 1
-        mbsecond <- nthArg 2 `mplus` (fmap (+ first) <$> namedArg "count")
-        step <- namedArg "step" `mplus` pure 1
+        mbsecond <- nthArg 2
+        step <- namedArg "step" 1
         pure $
           VArray $
             V.fromList $
@@ -375,9 +393,40 @@ construct =
               )
     ),
     ( "str",
-      makeFunction $ do
+      makeFunctionWithScope
+      (do
         val <- nthArg 1
-        VString <$> (fromVal val `mplus` pure (repr val))
+        base <- namedArg "base" (10 :: Integer)
+        let digitVector :: V.Vector Char
+            digitVector = V.fromList $ ['0'..'9'] ++ ['A'..'Z']
+        let renderDigit n = digitVector V.!? (fromIntegral n)
+        VString <$>
+          case val of
+            VInteger n | base /= 10
+              -> case mDigits base n of
+                   Nothing -> fail "Could not convert number to base"
+                   Just ds -> maybe
+                     (fail "Could not convert number to base")
+                     (pure . T.pack)
+                     (mapM renderDigit ds)
+            _ -> fromVal val `mplus` pure (repr val))
+      [ ( "to-unicode",
+           makeFunction $ do
+             (val :: Text) <- nthArg 1
+             case T.uncons val of
+               Just (c, t) | T.null t ->
+                 pure $ VInteger $ fromIntegral $ ord c
+               _ -> fail "to-unicode expects a single character" )
+      , ( "from-unicode",
+           makeFunction $ do
+             (val :: Int) <- nthArg 1
+             pure $ VString $ T.pack [chr val] )
+      ]
+    ),
+    ( "version",
+        makeFunction $ do
+          xs <- asks positional >>= mapM fromVal
+          pure $ VVersion xs
     ),
     ( "symbol",
       makeFunction $ do
@@ -436,13 +485,42 @@ hexToRGB _ = fail "expected string"
 
 loadFileLazyBytes :: Monad m => FilePath -> MP m BL.ByteString
 loadFileLazyBytes fp = do
-  loadBytes <- evalLoadBytes <$> getState
-  lift $ BL.fromStrict <$> loadBytes fp
+  operations <- evalOperations <$> getState
+  root <- evalPackageRoot <$> getState
+  lift $ BL.fromStrict <$> loadBytes operations (root </> fp)
 
 loadFileText :: Monad m => FilePath -> MP m T.Text
 loadFileText fp = do
-  loadBytes <- evalLoadBytes <$> getState
-  lift $ TE.decodeUtf8 <$> loadBytes fp
+  operations <- evalOperations <$> getState
+  root <- evalPackageRoot <$> getState
+  lift $ TE.decodeUtf8 <$> loadBytes operations (root </> fp)
+
+getUTCTime :: Monad m => MP m UTCTime
+getUTCTime = (currentUTCTime . evalOperations <$> getState) >>= lift
+
+time :: [(Identifier, Val)]
+time =
+  [ ( "datetime", makeFunctionWithScope
+      (do
+         mbyear <- namedArg "year" Nothing
+         mbmonth <- namedArg "month" Nothing
+         mbday <- namedArg "day" Nothing
+         let mbdate = case (mbyear, mbmonth, mbday) of
+                        (Just yr, Just mo, Just da) -> fromGregorianValid yr mo da
+                        _ -> Nothing
+         mbhour <- namedArg "hour" Nothing
+         mbminute <- namedArg "minute" Nothing
+         mbsecond <- namedArg "second" Nothing
+         let mbtime = case (mbhour, mbminute, mbsecond) of
+                        (Just hr, Just mi, Just se) ->
+                          Just $ secondsToDiffTime $ (hr * 60 * 60) + (mi * 60) + se
+                        _ -> Nothing
+         pure $ VDateTime mbdate mbtime)
+      [ ("today", makeFunction $ do
+            utcTime <- lift getUTCTime
+            pure $ VDateTime (Just (utctDay utcTime)) (Just (utctDayTime utcTime)) ) ]
+     )
+  ]
 
 dataLoading :: [(Identifier, Val)]
 dataLoading =
@@ -477,7 +555,14 @@ dataLoading =
         t <- lift $ loadFileText fp
         pure $ VString t
     ),
-    ("toml", makeFunction $ fail "unimplemented toml"),
+    ( "toml",
+      makeFunction $ do
+        fp <- nthArg 1
+        t <- lift $ loadFileText fp
+        case Toml.decode (T.unpack t) of
+          Toml.Failure e -> fail (unlines ("toml errors:" : e))
+          Toml.Success _ v -> pure v
+    ),
     ( "xml",
       makeFunction $ do
         fp <- nthArg 1
@@ -515,3 +600,40 @@ dataLoading =
                     ]
     )
   ]
+
+applyPureFunction :: Function -> [Val] -> Attempt Val
+applyPureFunction (Function f) vals =
+  let args = Arguments vals OM.empty
+   in case runParserT (f args) initialEvalState "" [] of
+        Failure s -> Failure s
+        Success (Left s) -> Failure $ show s
+        Success (Right v) -> Success v
+
+initialEvalState :: MonadFail m => EvalState m
+initialEvalState =
+  emptyEvalState { evalIdentifiers = [(BlockScope, standardModule)] }
+
+-- mDigitsRev, mDigits from the unmaintained digits package
+-- https://hackage.haskell.org/package/digits-0.3.1
+-- (c) 2009-2016 Henry Bucklow, Charlie Harvey -- BSD-3-Clause license.
+mDigitsRev :: Integral n
+    => n         -- ^ The base to use.
+    -> n         -- ^ The number to convert to digit form.
+    -> Maybe [n] -- ^ Nothing or Just the digits of the number in list form, in reverse.
+mDigitsRev base i = if base < 1
+                    then Nothing -- We do not support zero or negative bases
+                    else Just $ dr base i
+    where
+      dr _ 0 = []
+      dr b x = case base of
+                1 -> genericTake x $ repeat 1
+                _ -> let (rest, lastDigit) = quotRem x b
+                     in lastDigit : dr b rest
+
+-- | Returns the digits of a positive integer as a Maybe list.
+--   or Nothing if a zero or negative base is given
+mDigits :: Integral n
+    => n -- ^ The base to use.
+    -> n -- ^ The number to convert to digit form.
+    -> Maybe [n] -- ^ Nothing or Just the digits of the number in list form
+mDigits base i = reverse <$> mDigitsRev base i
