@@ -27,6 +27,8 @@ module Typst.Types
     MP,
     Scope (..),
     FlowDirective (..),
+    Operations (..),
+    XdgDirectory (..),
     EvalState (..),
     emptyEvalState,
     ShowRule (..),
@@ -43,6 +45,7 @@ module Typst.Types
     joinVals,
     prettyVal,
     valToContent,
+    prettyType,
     repr,
     Attempt (..),
   )
@@ -57,7 +60,7 @@ import qualified Data.Foldable as F
 import Data.Functor.Classes (Ord1 (liftCompare))
 import qualified Data.Map as M
 import qualified Data.Map.Ordered as OM
-import Data.Maybe (fromMaybe, isJust)
+import Data.Maybe (fromMaybe, isJust, catMaybes)
 import Data.Scientific (floatingOrInteger)
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
@@ -68,10 +71,16 @@ import qualified Data.Text as T
 import Data.Vector (Vector)
 import qualified Data.Vector as V
 import Text.Parsec
+import qualified Toml
+import qualified Toml.FromValue as Toml
+import qualified Toml.Pretty as Toml
 import qualified Text.PrettyPrint as P
 import Text.Read (readMaybe)
 import Typst.Regex (RE, makeLiteralRE)
 import Typst.Syntax (Identifier (..), Markup)
+import Data.Time (UTCTime, Day, DiffTime, timeOfDayToTime, localDay, localTimeOfDay)
+import Data.Time.Format (defaultTimeLocale, formatTime)
+import System.Directory (XdgDirectory(..))
 
 data Val
   = VNone
@@ -88,6 +97,7 @@ data Val
   | VSymbol !Symbol
   | VString !Text
   | VRegex !RE
+  | VDateTime (Maybe Day) (Maybe DiffTime)
   | VContent (Seq Content)
   | VArray (Vector Val)
   | VDict (OM.OMap Identifier Val)
@@ -102,6 +112,8 @@ data Val
   | VSelector !Selector
   | VModule Identifier (M.Map Identifier Val)
   | VStyles -- just a placeholder for now
+  | VVersion [Integer]
+  | VType !ValType
   deriving (Show, Eq, Typeable)
 
 instance FromJSON Val where
@@ -114,6 +126,22 @@ instance FromJSON Val where
   parseJSON (Aeson.Bool b) = pure $ VBoolean b
   parseJSON Aeson.Null = pure VNone
 
+instance Toml.FromValue Val where
+  fromValue = pure . tomlToVal
+
+tomlToVal :: Toml.Value -> Val
+tomlToVal (Toml.Bool x) = VBoolean x
+tomlToVal (Toml.Integer x) = VInteger x
+tomlToVal (Toml.String x) = VString (T.pack x)
+tomlToVal (Toml.Float x) = VFloat x
+tomlToVal (Toml.TimeOfDay x) = VDateTime Nothing (Just (timeOfDayToTime x))
+tomlToVal (Toml.Day x) = VDateTime (Just x) Nothing
+tomlToVal (Toml.LocalTime x) = VDateTime (Just (localDay x)) (Just (timeOfDayToTime (localTimeOfDay x)))
+tomlToVal (Toml.Array x) = VArray (V.fromList (map tomlToVal x))
+tomlToVal (Toml.Table x) = VDict (OM.fromList [(Identifier (T.pack k), tomlToVal v) | (k,v) <- M.assocs x])
+  -- typst specifies that unsupported datetimes map to strings and we don't have a place for the timezone
+tomlToVal v@Toml.ZonedTime{} = VString (T.pack (show (Toml.prettyValue v)))
+
 data ValType
   = TNone
   | TAuto
@@ -129,6 +157,7 @@ data ValType
   | TSymbol
   | TString
   | TRegex
+  | TDateTime
   | TContent
   | TArray
   | TDict
@@ -142,9 +171,11 @@ data ValType
   | TLabel
   | TCounter
   | TLocation
+  | TVersion
+  | TType
   | TAny
   | ValType :|: ValType
-  deriving (Show, Eq, Typeable)
+  deriving (Show, Eq, Ord, Typeable)
 
 valType :: Val -> ValType
 valType v =
@@ -163,6 +194,7 @@ valType v =
     VSymbol {} -> TSymbol
     VString {} -> TString
     VRegex {} -> TRegex
+    VDateTime {} -> TDateTime
     VContent {} -> TContent
     VArray {} -> TArray
     VDict {} -> TDict
@@ -175,6 +207,8 @@ valType v =
     VModule {} -> TModule
     VSelector {} -> TSelector
     VStyles {} -> TStyles
+    VVersion {} -> TVersion
+    VType {} -> TType
 
 hasType :: ValType -> Val -> Bool
 hasType TAny _ = True
@@ -309,7 +343,7 @@ joinVals = go
     go (VContent cs) (VString t) = pure $ VContent (cs Seq.|> Txt t)
     go (VContent cs) (VContent cs') = pure $ VContent (cs <> cs')
     go (VArray vec) (VArray vec') = pure $ VArray (vec <> vec')
-    go accum v = fail $ "Can't combine " <> show accum <> " and " <> show v
+    go x y = pure $ VContent $ valToContent x <> valToContent y
 
 class Compare a where
   comp :: a -> a -> Maybe Ordering
@@ -340,6 +374,16 @@ instance Compare Val where
   comp (VDict m1) (VDict m2) =
     Just $ liftCompare (\x y -> fromMaybe LT (comp x y)) (OM.toMap m1) (OM.toMap m2)
   comp (VFunction (Just i1) _ _) (VFunction (Just i2) _ _) = Just $ compare i1 i2
+  comp (VType ty1) (VType ty2) = Just $ compare ty1 ty2
+  comp (VType TInteger) (VString "integer") = Just EQ
+  comp (VString "integer") (VType TInteger) = Just EQ
+  comp (VType ty) (VString s) = Just $ compare (prettyType ty) s
+  comp (VString s) (VType ty)  = Just $ compare s (prettyType ty)
+  comp (VVersion as) (VVersion bs)
+    | length as > length bs =
+       Just $ compare as (bs ++ replicate (length as - length bs) 0)
+    | otherwise =
+       Just $ compare (as ++ replicate (length as - length bs) 0) bs
   comp _ _ = Nothing
 
 instance Ord Val where
@@ -378,6 +422,8 @@ instance Summable Val where
   maybePlus (VContent c1) (VContent c2) = pure $ VContent (c1 <> c2)
   maybePlus (VString s1) (VContent c2) = pure $ VContent (Txt s1 Seq.<| c2)
   maybePlus (VContent c1) (VString s2) = pure $ VContent (c1 Seq.|> Txt s2)
+  maybePlus (VString s1) sym@(VSymbol{}) = pure $ VString (s1 <> repr sym)
+  maybePlus sym@(VSymbol{}) (VString s2) = pure $ VString (repr sym <> s2)
   maybePlus (VLength l1) (VLength l2) = pure $ VLength (l1 <> l2)
   maybePlus (VLength l1) (VRatio r1) = pure $ VLength (l1 <> LRatio r1)
   maybePlus (VRatio r1) (VLength l1) = pure $ VLength (l1 <> LRatio r1)
@@ -414,6 +460,8 @@ instance Multipliable Val where
     | i >= 0 = pure $ VContent (mconcat $ replicate (fromIntegral i) c)
   maybeTimes (VInteger i) (VLength l) = pure $ VLength $ timesLength (fromIntegral i) l
   maybeTimes (VLength l) (VInteger i) = pure $ VLength $ timesLength (fromIntegral i) l
+  maybeTimes (VRatio r) (VLength l) = pure $ VLength $ timesLength (fromRational r) l
+  maybeTimes (VLength l) (VRatio r) = pure $ VLength $ timesLength (fromRational r) l
   maybeTimes (VFloat f) (VLength l) = pure $ VLength $ timesLength f l
   maybeTimes (VLength l) (VFloat f) = pure $ VLength $ timesLength f l
   maybeTimes (VInteger i) (VAngle a) = pure $ VAngle (fromIntegral i * a)
@@ -515,6 +563,14 @@ data FlowDirective
   | FlowReturn Bool
   deriving (Show, Ord, Eq)
 
+data Operations m =
+  Operations
+  { loadBytes :: FilePath -> m BS.ByteString
+  , currentUTCTime :: m UTCTime
+  , lookupEnvVar :: String -> m (Maybe String)
+  , checkExistence :: FilePath -> m Bool
+  }
+
 data EvalState m = EvalState
   { evalIdentifiers :: [(Scope, M.Map Identifier Val)],
     -- first item is current block, then superordinate block, etc.
@@ -523,7 +579,8 @@ data EvalState m = EvalState
     evalShowRules :: [ShowRule],
     evalStyles :: M.Map Identifier Arguments,
     evalFlowDirective :: FlowDirective,
-    evalLoadBytes :: FilePath -> m BS.ByteString
+    evalPackageRoot :: FilePath,
+    evalOperations :: Operations m
   }
 
 emptyEvalState :: EvalState m
@@ -534,7 +591,8 @@ emptyEvalState = EvalState
       evalShowRules = [],
       evalStyles = mempty,
       evalFlowDirective = FlowNormal,
-      evalLoadBytes = undefined
+      evalPackageRoot = mempty,
+      evalOperations = undefined
     }
 
 data Attempt a
@@ -719,6 +777,8 @@ prettyVal expr =
     VContent cs -> prettyContent cs
     VString t -> "\"" <> escString t <> "\""
     VRegex re -> P.text (show re)
+    VDateTime d t -> P.text (unwords (catMaybes
+       [show <$> d, formatTime defaultTimeLocale "%0H:%0M:%0S" <$> t]))
     VAuto -> "auto"
     VNone -> "none"
     VBoolean True -> "true"
@@ -759,7 +819,7 @@ prettyVal expr =
         )
     VDirection d -> text $ T.toLower $ T.pack $ show d
     VFunction _ _ _ -> mempty
-    VLabel _ -> mempty
+    VLabel t -> text $ "<" <> t <> ">"
     VCounter _ -> mempty
     VColor (RGB r g b o) ->
       "rgb("
@@ -809,6 +869,13 @@ prettyVal expr =
     VSymbol (Symbol t _ _) -> text t
     VSelector _ -> mempty
     VStyles -> mempty
+    VVersion xs -> text $ T.intercalate "." (map (T.pack . show) xs)
+    VType ty -> text $ prettyType ty
+
+prettyType :: ValType -> Text
+prettyType TDict = "dictionary"
+prettyType TInteger = "int"
+prettyType x = T.toLower . T.pack . drop 1 . show $ x
 
 escString :: Text -> P.Doc
 escString =
